# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
File included to make the directory a Python package.

The test_*.py files are special in this directory in that they refer to the
top level module names of the adapter to test. An attempt will be made to
import that module but ignored if it fails (not available on the system).
"""
